#!/usr/bin/env python3
#
# image crawler
#
# the image crawler checks for new openstack ("cloud") images
# whenever a new image is detected all relevant information needed for
# maintaining an image catalog
#
# 2022-09-19 v0.1 christian.stelter@plusserver.com

import argparse
import sys
import os

from crawler.core.config import config_read
from crawler.core.database import (
    database_connect,
    database_disconnect,
    database_initialize,
)
from crawler.core.exporter import export_image_catalog, export_image_catalog_all
from crawler.core.main import crawl_image_sources
from crawler.git.base import clone_or_pull, update_repository


def main():
    print("\nplusserver Image Crawler v0.1\n")

    working_directory = os.getcwd()

    parser = argparse.ArgumentParser(
        description="checks cloud image repositories for new updates and"
        + " keeps track of all images within its sqlite3 database"
    )
    parser.add_argument(
        "--config",
        type=str,
        required=False,
        help="specify the config file to be used (default: etc/config.yaml)",
    )
    parser.add_argument(
        "--sources",
        type=str,
        required=False,
        help="specify the sources file to be used - overrides value from config file",
    )
    parser.add_argument(
        "--init-db",
        action="store_true",
        required=False,
        help="initialize image catalog database",
    )
    parser.add_argument(
        "--export-only",
        action="store_true",
        required=False,
        help="export only existing image catalog",
    )
    parser.add_argument(
        "--updates-only",
        action="store_true",
        required=False,
        help="check only for updates, do not export catalog",
    )
    args = parser.parse_args()

    # read configuration
    if args.config is not None:
        config_filename = args.config
    else:
        # default
        config_filename = "etc/config.yaml"

    config = config_read(config_filename, "configuration")
    if config is None:
        raise SystemExit("\nERROR: Unable to open config " + config_filename)

    # read the image sources
    if args.sources is not None:
        sources_filename = args.sources
    else:
        sources_filename = config["sources_name"]

    image_source_catalog = config_read(sources_filename, "source catalog")
    if image_source_catalog is None:
        raise SystemExit("Unable to open image source catalog " + sources_filename)

    # initialize database when run with --init-db
    if args.init_db:
        database_initialize(config["database_name"])
        sys.exit(0)

    # clone or update local repository when git is enabled
    if "remote_repository" in config:
        if "git_ssh_command" in config:
            git_ssh_command = config["git_ssh_command"]
        else:
            git_ssh_command = None
        if "branch" in config:
            working_branch = config["branch"]
        else:
            working_branch = main
        clone_or_pull(
            config["remote_repository"],
            config["local_repository"],
            working_branch,
            git_ssh_command,
        )
    else:
        print("No image catalog repository configured")

    # connect to database
    database = database_connect(config["database_name"])
    if database is None:
        print("\nERROR: Could not open database %s" % config["database_name"])
        print(
            '\nRun "./image-crawler.py --init-db" to create a new database OR config check your etc/config.yaml'
        )
        sys.exit(1)

    # crawl image sources when requested
    if args.export_only:
        print("\nSkipping repository crawling")
        updated_sources = {}
    else:
        print("\nStart repository crawling")
        updated_sources = crawl_image_sources(image_source_catalog, database)

    # export image catalog
    if args.updates_only:
        print("\nSkipping catalog export")
    else:
        if config["local_repository"].startswith("/"):
            export_path = config["local_repository"]
        else:
            export_path = working_directory + "/" + config["local_repository"]

        if updated_sources:
            print("\nExporting catalog to %s" % export_path)
            export_image_catalog(
                database,
                image_source_catalog,
                updated_sources,
                config["local_repository"],
                config["template_path"],
            )
        else:
            if args.export_only:
                print("\nExporting all catalog files to %s" % export_path)
                export_image_catalog_all(
                    database,
                    image_source_catalog,
                    config["local_repository"],
                    config["template_path"],
                )

    # push changes to git repository when configured
    if "remote_repository" in config and updated_sources:
        update_repository(
            database, config["local_repository"], updated_sources, git_ssh_command
        )
    else:
        print("No remote repository update needed.")

    database_disconnect(database)


if __name__ == "__main__":
    main()



import requests
import validators

from email.utils import parsedate_to_datetime

# from pprint import pprint


def url_fetch_content(url):
    try:
        request = requests.get(url, allow_redirects=True)
    except requests.exceptions.HTTPError as errh:
        print("ERROR: HTTP error: " + repr(errh))
        return None
    except requests.exceptions.ConnectionError as errc:
        print("ERROR: could not connect to the API: " + repr(errc))
        return None
    except requests.exceptions.Timeout as errt:
        # is not hit when timeout - why?
        print("ERROR: timeout while connecting: " + repr(errt))
        return None
    except requests.exceptions.RequestException as err:
        print("ERROR: unknown error: " + repr(err))
        return None

    content = request.content.decode("utf-8")

    return content


def url_get_header(url):
    if not validators.url(url):
        print("ERROR: ", url, " is not a valid URL")
        return None

    try:
        request = requests.head(url, allow_redirects=True)
        # info.raise_for_status()
    except requests.exceptions.HTTPError as errh:
        print("ERROR: HTTP error: " + repr(errh))
        return None
    except requests.exceptions.ConnectionError as errc:
        print("ERROR: could not connect to the API: " + repr(errc))
        return None
    except requests.exceptions.Timeout as errt:
        # is not hit when timeout - why?
        print("ERROR: timeout while connecting: " + repr(errt))
        return None
    except requests.exceptions.RequestException as err:
        print("ERROR: unknown error: " + repr(err))
        return None

    if request.status_code == 404:
        return None

    return request


def url_get_last_modified(url):
    request = url_get_header(url)
    if request is None:
        return None

    last_modified_date = request.headers["Last-Modified"]
    datestring = parsedate_to_datetime(last_modified_date).strftime("%Y-%m-%d")

    return datestring


def url_exists(url):
    request = url_get_header(url)

    if request is None:
        return False
    else:
        return True



from jinja2 import Template
import os

from crawler.core.database import read_release_from_catalog


def export_image_catalog(
    connection, sources_catalog, updated_sources, local_repository, template_path
):
    # "smart" export - only releases with updates will be written
    # create directory (once) - only necessary when not created by git clone
    if not os.path.exists(local_repository):
        try:
            print("Creating repository directory (%s)" % local_repository)
            os.makedirs(local_repository)
        except os.error as error:
            raise SystemExit(
                "FATAL: Creating directory %s failed with %s"
                % (local_repository, error)
            )

    for source in sources_catalog["sources"]:
        if source["name"] in updated_sources:
            distribution = source["name"]
            print("Exporting image catalog for " + distribution)
            header_file = open(template_path + "/header.yml")
            catalog_export = header_file.read()
            header_file.close()

            image_template_filename = (
                template_path + "/" + distribution.lower().replace(" ", "_") + ".yml.j2"
            )
            image_template_file = open(image_template_filename, "r")
            image_template = Template(image_template_file.read())
            image_template_file.close()

            for release in source["releases"]:
                # TODO check empty catalog (still necessary after add updated sources?)
                release_catalog = read_release_from_catalog(
                    connection, distribution, release["name"]
                )
                release_catalog["name"] = distribution
                release_catalog["os_distro"] = distribution.lower()
                release_catalog["os_version"] = release["name"]
                release_catalog["codename"] = release["codename"]

                catalog_export = (
                    catalog_export
                    + image_template.render(catalog=release_catalog, metadata=release)
                    + "\n"
                )

            image_catalog_export_filename = (
                local_repository + "/" + distribution.lower().replace(" ", "_") + ".yml"
            )
            # TODO error handling
            image_catalog_export_file = open(image_catalog_export_filename, "w")
            image_catalog_export_file.write(catalog_export)
            image_catalog_export_file.close()


def export_image_catalog_all(
    connection, sources_catalog, local_repository, template_path
):
    # export all releases - used with --export-only
    # create directory (once) - only necessary when not created by git clone
    if not os.path.exists(local_repository):
        try:
            print("Creating repository directory (%s)" % local_repository)
            os.makedirs(local_repository)
        except os.error as error:
            raise SystemExit(
                "FATAL: Creating directory %s failed with %s"
                % (local_repository, error)
            )

    for source in sources_catalog["sources"]:
        distribution = source["name"]
        print("Exporting image catalog for " + distribution)
        header_file = open(template_path + "/header.yml")
        catalog_export = header_file.read()
        header_file.close()

        image_template_filename = (
            template_path + "/" + distribution.lower().replace(" ", "_") + ".yml.j2"
        )
        image_template_file = open(image_template_filename, "r")
        image_template = Template(image_template_file.read())
        image_template_file.close()

        for release in source["releases"]:
            # TODO check empty catalog (still necessary after add updated sources?)
            release_catalog = read_release_from_catalog(
                connection, distribution, release["name"]
            )
            release_catalog["name"] = distribution
            release_catalog["os_distro"] = distribution.lower()
            release_catalog["os_version"] = release["name"]
            release_catalog["codename"] = release["codename"]

            catalog_export = (
                catalog_export
                + image_template.render(catalog=release_catalog, metadata=release)
                + "\n"
            )

        image_catalog_export_filename = (
            local_repository + "/" + distribution.lower().replace(" ", "_") + ".yml"
        )
        # TODO error handling
        image_catalog_export_file = open(image_catalog_export_filename, "w")
        image_catalog_export_file.write(catalog_export)
        image_catalog_export_file.close()



import yaml
from pathlib import Path


def config_read(name, msg="config"):
    path = Path(name)
    if not path.is_file():
        return None

    try:
        config = yaml.safe_load(Path(name).read_text())
    except PermissionError:
        print("ERROR: could not open config - please check file permissions")
        return None
    except yaml.YAMLError as error:
        print("ERROR: %s" % error)
        return None

    print("Successfully read %s from %s" % (msg, name))

    return config



import requests
import datetime

from bs4 import BeautifulSoup

from crawler.web.generic import url_fetch_content


def web_get_checksum(url, imagename):
    checksum_list = url_fetch_content(url)
    if checksum_list is None:
        return None

    for line in checksum_list.splitlines():
        if imagename in line:
            (new_checksum, filename) = line.split()
            return new_checksum

    return None


def release_build_image_url(release, versionpath, version):
    # works for Ubuntu, Debian
    if not release["baseURL"].endswith("/"):
        base_url = release["baseURL"] + "/"
    else:
        base_url = release["baseURL"]

    if "debian" in release["imagename"]:
        return (
            base_url
            + versionpath
            + release["imagename"]
            + "-"
            + version
            + "."
            + release["extension"]
        )
    elif "ubuntu" in release["imagename"]:
        return (
            base_url + versionpath + release["imagename"] + "." + release["extension"]
        )
    elif "AlmaLinux" in release["imagename"]:
        return base_url + release["releasepath"] + "/" + versionpath
    # we do not know this distribution
    else:
        return None


def release_get_version_from_path(release, versionpath):
    if "debian" in release["imagename"]:
        if versionpath.endswith("/"):
            versionpath = versionpath.rstrip("/")
        return versionpath
    elif "ubuntu" in release["imagename"]:
        if versionpath.endswith("/"):
            versionpath = versionpath.rstrip("/")
        return versionpath.replace("release-", "")
    # we do not know this distribution
    else:
        return None


def release_get_release_from_path(release, versionpath):
    if "debian" in release["imagename"]:
        release_date = (
            versionpath[0:4] + "-" + versionpath[4:6] + "-" + versionpath[6:8]
        )
        return release_date
    elif "ubuntu" in release["imagename"]:
        release_version = versionpath.replace("release-", "")
        release_date = (
            release_version[0:4]
            + "-"
            + release_version[4:6]
            + "-"
            + release_version[6:8]
        )
        return release_date
    # we do not know this distribution
    else:
        return None


def web_get_current_image_metadata(release, image_filedate):
    # optimistic release date assumption as in release date == upload date
    version = image_filedate.replace("-", "")

    # get directory content
    if "debian" in release["imagename"] or "ubuntu" in release["imagename"]:
        requestURL = release["baseURL"]
    else:
        requestURL = release["baseURL"] + release["releasepath"]

    request = requests.get(requestURL, allow_redirects=True)
    soup = BeautifulSoup(request.text, "html.parser")

    for link in soup.find_all("a"):
        data = link.get("href")
        if data.find(version) != -1:
            release_version_path = data
            if "debian" in release["imagename"] or "ubuntu" in release["imagename"]:
                new_version = release_get_version_from_path(
                    release, release_version_path
                )
                if new_version is None:
                    return None
            else:
                new_version = version
            # check image URL?

            if "debian" in release["imagename"] or "ubuntu" in release["imagename"]:
                release_date = release_get_release_from_path(
                    release, release_version_path
                )
                if release_date is None:
                    return None
            else:
                release_date = image_filedate

            return {
                "url": release_build_image_url(
                    release, release_version_path, new_version
                ),
                "version": new_version,
                "release_date": release_date,
            }

    # release is behind file date
    filedate = datetime.date(int(version[0:4]), int(version[4:6]), int(version[6:8]))

    max_days_back = 6
    days_back = 1

    while days_back <= max_days_back:
        filedate = filedate - datetime.timedelta(days=1)
        new_version = filedate.strftime("%Y%m%d")

        for link in soup.find_all("a"):
            data = link.get("href")
            if data.find(new_version) != -1:
                release_version_path = data
                if release_version_path.endswith("/"):
                    release_version_path = release_version_path.rstrip("/")
                new_version = release_get_version_from_path(
                    release, release_version_path
                )

                # RELEASE DATE ?
                if new_version is None:
                    return None
                # check image URL?

                release_date = release_get_release_from_path(
                    release, release_version_path
                )
                if release_date is None:
                    return None

                return {
                    "url": release_build_image_url(
                        release, release_version_path, new_version
                    ),
                    "version": new_version,
                    "release_date": release_date,
                }

        days_back = days_back + 1

    return None



import sys
import sqlite3
from pathlib import Path


def database_connect(name):
    path = Path(name)
    if path.is_file():
        try:
            connection = sqlite3.connect(name)
        except sqlite3.OperationalError as error:
            print("ERROR: %s" % error)
            return None
        return connection
    else:
        return None


def database_disconnect(connection):
    connection.close()


def database_initialize(name):
    path = Path(name)
    if path.is_file():
        print("WARNING: database %s already exists. Cowardly refusing action." % name)
    else:
        create_statement_file_path = Path("lib/initialize-image-catalog.sql")
        if create_statement_file_path.is_file():
            db_init_file = open("lib/initialize-image-catalog.sql")
            create_statement = db_init_file.read()
            db_init_file.close()
        else:
            raise SystemError("Template lib/initialize-image-catalog.sql not found")

        try:
            connection = sqlite3.connect(name)
        except sqlite3.OperationalError as error:
            print("ERROR: %s" % error)
        database_cursor = connection.cursor()
        try:
            database_cursor.execute(create_statement)
        except Exception as error:
            print('ERROR: create table failed with the following error "%s"' % error)

        connection.close()
        print("New database created under %s" % name)


def db_get_last_checksum(connection, distribution, release):
    try:
        database_cursor = connection.cursor()
        database_cursor.execute(
            "SELECT checksum FROM image_catalog "
            "WHERE distribution_name = '%s' "
            "AND distribution_release = '%s' "
            "ORDER BY id DESC LIMIT 1" % (distribution, release)
        )
    except sqlite3.OperationalError as error:
        raise SystemError("SQLite error: %s" % error)

    row = database_cursor.fetchone()

    if row is None:
        # print("no previous entries found")
        last_checksum = "sha256:none"
    else:
        last_checksum = row[0]

    database_cursor.close()

    return last_checksum


def db_get_last_entry(connection, distribution, release):
    return db_get_release_versions(connection, distribution, release, 1)


def db_get_release_versions(connection, distribution, release, limit):
    try:
        database_cursor = connection.cursor()
        database_cursor.execute(
            "SELECT name, release_date, version, distribution_name, "
            "distribution_release, url, checksum FROM image_catalog "
            "WHERE distribution_name = '%s' AND distribution_release = '%s' "
            "ORDER BY id DESC LIMIT %d" % (distribution, release, limit)
        )
    except sqlite3.OperationalError as error:
        print("SQLite error: %s" % error)
        sys.exit(1)
    row = database_cursor.fetchone()

    if row is not None:
        last_entry = {}
        last_entry["name"] = row[0]
        last_entry["release_date"] = row[1]
        last_entry["version"] = row[2]
        last_entry["distribution_name"] = row[3]
        last_entry["distribution_version"] = row[4]
        last_entry["url"] = row[5]
        last_entry["checksum"] = row[6]

    database_cursor.close()

    return last_entry


def read_version_from_catalog(connection, distribution, release, version):
    try:
        database_cursor = connection.cursor()
        database_cursor.execute(
            "SELECT version,checksum,url,release_date "
            "FROM (SELECT * FROM image_catalog "
            "WHERE distribution_name = '%s' "
            "AND distribution_release = '%s' "
            "AND version ='%s' "
            "ORDER BY id DESC LIMIT 1) "
            "ORDER BY ID" % (distribution, release, version)
        )
    except sqlite3.OperationalError as error:
        raise SystemError("SQLite error: %s" % error)

    image_catalog = {}
    image_catalog["versions"] = {}

    for image in database_cursor.fetchall():
        version = image[0]
        image_catalog["versions"][version] = {}
        image_catalog["versions"][version]["checksum"] = image[1]
        image_catalog["versions"][version]["url"] = image[2]
        image_catalog["versions"][version]["release_date"] = image[3]

    return image_catalog


def write_catalog_entry(connection, update):
    try:
        database_cursor = connection.cursor()
        database_cursor.execute(
            "INSERT INTO image_catalog "
            "(name, release_date, version, distribution_name, "
            "distribution_release, url, checksum) "
            "VALUES (?,?,?,?,?,?,?)",
            (
                update["name"],
                update["release_date"],
                update["version"],
                update["distribution_name"],
                update["distribution_release"],
                update["url"],
                update["checksum"],
            ),
        )
        connection.commit()
    except sqlite3.OperationalError as error:
        raise SystemError("SQLite error: %s" % error)

    database_cursor.close()

    return None


def update_catalog_entry(connection, update):
    try:
        database_cursor = connection.cursor()
        database_cursor.execute(
            "UPDATE image_catalog set url=?, checksum=? " "WHERE name=? AND version=?",
            (update["url"], update["checksum"], update["name"], update["version"]),
        )
        connection.commit()
    except sqlite3.OperationalError as error:
        raise SystemError("SQLite error: %s" % error)

    database_cursor.close()

    return None


def write_or_update_catalog_entry(connection, update):
    existing_entry = read_version_from_catalog(
        connection,
        update["distribution_name"],
        update["distribution_release"],
        update["version"],
    )

    if update["version"] in existing_entry["versions"]:
        print("Updating version " + update["version"])
        return update_catalog_entry(connection, update)
    else:
        return write_catalog_entry(connection, update)


def read_release_from_catalog(connection, distribution, release):
    try:
        database_cursor = connection.cursor()
        database_cursor.execute(
            "SELECT version,checksum,url,release_date "
            "FROM (SELECT * FROM image_catalog "
            "WHERE distribution_name = '%s' "
            "AND distribution_release = '%s' "
            "ORDER BY id DESC LIMIT 3) "
            "ORDER BY ID" % (distribution, release)
        )
    except sqlite3.OperationalError as error:
        raise SystemError("SQLite error: %s" % error)

    image_catalog = {}
    image_catalog["versions"] = {}

    for image in database_cursor.fetchall():
        version = image[0]
        image_catalog["versions"][version] = {}
        image_catalog["versions"][version]["checksum"] = image[1]
        image_catalog["versions"][version]["url"] = image[2]
        image_catalog["versions"][version]["release_date"] = image[3]

    return image_catalog



from crawler.core.database import db_get_last_checksum, write_or_update_catalog_entry
from crawler.web.generic import url_get_last_modified
from crawler.web.directory import web_get_checksum, web_get_current_image_metadata


def release_update_check(release, last_checksum):
    # works for Ubuntu, Debian
    if not release["baseURL"].endswith("/"):
        base_url = release["baseURL"] + "/"
    else:
        base_url = release["baseURL"]

    # check on leading an trialing slash / for release path ?
    checksum_url = base_url + release["releasepath"] + "/" + release["checksumname"]
    # works for Ubuntu, Debian
    # imagename _with_ proper extension to look for in checksum lists
    imagename = release["imagename"] + "." + release["extension"]

    current_checksum = web_get_checksum(checksum_url, imagename)
    if current_checksum is None:
        print(
            "ERROR: no matching checksum found - check image (%s) "
            "and checksum filename (%s)" % (imagename, release["checksumname"])
        )
        return None

    current_checksum = release["algorithm"] + ":" + current_checksum

    if current_checksum != last_checksum:
        image_url = base_url + release["releasepath"] + "/" + imagename
        image_filedate = url_get_last_modified(image_url)

        image_metadata = web_get_current_image_metadata(release, image_filedate)
        if image_metadata is not None:

            update = {}
            update["release_date"] = image_metadata["release_date"]
            update["url"] = image_metadata["url"]
            update["version"] = image_metadata["version"]
            update["checksum"] = current_checksum
            return update
        else:
            return None

    return None


def image_update_service(connection, source):
    updated_releases = []
    for release in source["releases"]:
        last_checksum = db_get_last_checksum(
            connection, source["name"], release["name"]
        )
        catalog_update = release_update_check(release, last_checksum)
        if catalog_update:
            print("Update found for " + source["name"] + " " + release["name"])
            print("New release " + catalog_update["version"])
            # catalog_update anreichern mit _allen_ Daten für die DB
            catalog_update["name"] = source["name"] + " " + release["name"]
            catalog_update["distribution_name"] = source["name"]
            catalog_update["distribution_release"] = release["name"]
            catalog_update["release"] = release["name"]

            write_or_update_catalog_entry(connection, catalog_update)
            updated_releases.append(release["name"])
        else:
            print("No update found for " + source["name"] + " " + release["name"])

    return updated_releases



#!/usr/bin/env python3
import sys
from crawler.web.generic import url_exists
from crawler.web.directory import web_get_checksum
from crawler.core.database import (
    database_connect,
    database_disconnect,
    write_catalog_entry,
)

import requests
from bs4 import BeautifulSoup
from pprint import pprint

from crawler.core.config import config_read

import re


def get_all_versions_with_paths_from_directory(release):
    version_paths = {}

    # get release directory content
    if "debian" in release["imagename"]:
        requestURL = release["baseURL"]
        version_regex = re.compile(r"(\d{8}-\d*)")
        path_regex = version_regex
    elif "ubuntu" in release["imagename"]:
        requestURL = release["baseURL"]
        version_regex = re.compile(r"release-(\d{8})\.*\d*")
        path_regex = re.compile(r"(release-\d{8}\.*\d*)")
    else:
        # not yet supported
        return version_paths
        # requestURL = release['baseURL'] + release['releasepath']
        # workaround for Alma
        # version_regex = re.compile(r'AlmaLinux-9-GenericCloud-9.0-(\d{8})')
        # path_regex = re.compile(r'AlmaLinux-9-GenericCloud-9.0-(\d{8})')

    request = requests.get(requestURL, allow_redirects=True)
    soup = BeautifulSoup(request.text, "html.parser")

    last_version = ""
    for link in soup.find_all("a"):
        data = link.get("href")
        # print(data)
        match = version_regex.search(data)
        if match is not None:
            # just use first occurance of matching pattern
            # Debian has two links for one version
            if match.group(1) != last_version:
                version = match.group(1)
                version_paths[version] = {}
                path_match = path_regex.search(data)
                version_paths[version]["path"] = path_match.group(1)
                last_version = version

    return version_paths


# only Debian and Ubuntu?
def get_version_path(release, version):

    pprint(version)
    pprint(release)
    # get release directory content
    if "debian" in release["imagename"]:
        requestURL = release["baseURL"]
        version_regex = re.compile(r"\d8-\d*")
    elif "ubuntu" in release["imagename"]:
        requestURL = release["baseURL"]
        version_regex = re.compile(r"release-\d{8}\.*\d*")
    # not yet supported
    # else:
    #     requestURL = release['baseURL'] + release['releasepath']
    #     version_regex = re.compile(r'AlmaLinux-9-GenericCloud-9.0-\d{8}')

    request = requests.get(requestURL, allow_redirects=True)
    soup = BeautifulSoup(request.text, "html.parser")

    for link in soup.find_all("a"):
        data = link.get("href")
        if data.find(version) != -1:
            match = version_regex.search(data)
            if match is not None:
                # just use first occurance of matching pattern
                # Debian has two links for one version
                path = match.group()
                return path

    return None


def get_checksum_from_version_path(release, version_path):
    if "debian" in release["imagename"]:
        url = release["baseURL"] + version_path + "/" + release["checksumname"]
        imagename_fq = (
            release["imagename"] + "-" + version_path + "." + release["extension"]
        )
    elif "ubuntu" in release["imagename"]:
        url = release["baseURL"] + version_path + "/" + release["checksumname"]
        imagename_fq = release["imagename"] + "." + release["extension"]
    else:
        # not yet supported
        return None

    # print(url)
    return web_get_checksum(url, imagename_fq)


def get_correct_version_path(release, version):
    # should be needed for ubuntu with its .x ending paths
    if "debian" in release["imagename"] or "ubuntu" in release["imagename"]:
        requestURL = release["baseURL"]
    else:
        requestURL = release["baseURL"] + release["releasepath"]

    request = requests.get(requestURL, allow_redirects=True)
    soup = BeautifulSoup(request.text, "html.parser")

    for link in soup.find_all("a"):
        data = link.get("href")
        if data.endswith("/"):
            path = data.replace("/", "")
        else:
            path = data

        if path.find(version) != -1:
            print("WARNING: Path correction necessary %s (or not found)" % path)
            return path

    return None


def get_version_metadata(release, version):
    metadata = {}

    # 1. check path
    if "debian" in release["imagename"]:
        url = (
            release["baseURL"]
            + version
            + "/"
            + release["imagename"]
            + "-"
            + version
            + "."
            + release["extension"]
        )
    elif "ubuntu" in release["imagename"]:
        url = (
            release["baseURL"]
            + "release-"
            + version
            + "/"
            + release["imagename"]
            + "."
            + release["extension"]
        )
    else:
        # not yet supported
        return metadata

    if url_exists(url):
        metadata["url"] = url
        if "debian" in release["imagename"]:
            version_path = version
        elif "ubuntu" in release["imagename"]:
            version_path = "release-" + version
    else:
        if "ubuntu" in release["imagename"]:
            version_path = get_correct_version_path(release, version)
            url = (
                release["baseURL"]
                + version_path
                + "/"
                + release["imagename"]
                + "."
                + release["extension"]
            )
            if url_exists(url):
                metadata["url"] = url
            else:
                print("WARNING %s does not exist" % url)
                return metadata
        else:
            print("WARNING %s does not exist" % url)
            return metadata

    # 2. get checksum
    checksum = get_checksum_from_version_path(release, version_path)
    if checksum is None:
        print("WARNING: Could not get checksum for version %s" % version)
        return {}

    metadata["checksum"] = release["algorithm"] + ":" + checksum

    # 3. extract release date from version
    release_date = version[0:4] + "-" + version[4:6] + "-" + version[6:8]
    metadata["release_date"] = release_date

    # 4. add version to metadata
    metadata["version"] = version

    return metadata


def main():
    # read configuration
    config_filename = "etc/config.yaml"

    config = config_read(config_filename, "configuration")
    if config is None:
        raise SystemExit("\nERROR: Unable to open config " + config_filename)

    # read the image sources
    sources_filename = config["sources_name"]

    image_source_catalog = config_read(sources_filename, "source catalog")
    if image_source_catalog is None:
        raise SystemExit("Unable to open image source catalog " + sources_filename)

    # connect to database
    database = database_connect(config["database_name"])
    if database is None:
        print("\nERROR: Could not open database %s" % config["database_name"])
        print(
            '\nRun "./image-crawler.py --init-db" to create a new database OR config check your etc/config.yaml'
        )
        sys.exit(1)

    for source in image_source_catalog["sources"]:
        for release in source["releases"]:
            # pprint(release)
            print("Crawling %s %s" % (source["name"], release["name"]))
            # print(get_all_versions_with_paths_from_directory(release))
            version_paths = get_all_versions_with_paths_from_directory(release)
            for version in version_paths:
                print("Getting metadata for version %s" % version)
                catalog_update = get_version_metadata(release, version)
                if not catalog_update:
                    print(
                        "WARNING: Skipping version %s due to missing metadata" % version
                    )
                else:
                    catalog_update["name"] = source["name"] + " " + release["name"]
                    catalog_update["distribution_name"] = source["name"]
                    catalog_update["distribution_release"] = release["name"]
                    catalog_update["release"] = release["name"]

                    write_catalog_entry(database, catalog_update)

    database_disconnect(database)


if __name__ == "__main__":
    main()



from crawler.updater.service import image_update_service


def crawl_image_sources(image_source_catalog, database):

    updated_sources = {}
    for source in image_source_catalog["sources"]:
        print("\nChecking updates for Distribution " + source["name"])
        updated_releases = image_update_service(database, source)
        if updated_releases:
            updated_sources[source["name"]] = {}
            updated_sources[source["name"]]["releases"] = updated_releases

    return updated_sources



import git
import os
from pathlib import Path

from crawler.core.database import db_get_release_versions


def clone_or_pull(remote_repository, repository, working_branch, ssh_command):
    if ssh_command:
        os.environ["GIT_SSH_COMMAND"] = ssh_command
    path = Path(repository)
    if not path.is_dir():
        print(
            "Cloning repository %s (%s) into %s"
            % (remote_repository, working_branch, repository)
        )
        try:
            image_repo = git.Repo.clone_from(
                remote_repository, repository, branch=working_branch
            )
        except git.exc.GitCommandError as error:
            raise SystemExit(
                "FATAL: Cloning of %s failed with %s" % (remote_repository, error)
            )
    else:
        print("Repository exists already, pulling changes (%s)" % working_branch)
        image_repo = git.Repo(repository)
        try:
            image_repo.remotes.origin.pull()
        except git.exc.GitCommandError as error:
            raise SystemExit("FATAL: Update (pull) failed with %s" % error)

        try:
            image_repo.git.checkout(working_branch)
        except git.exc.GitCommandError as error:
            raise SystemExit(
                "FATAL: Checkout on branch %s failed with %s" % (working_branch, error)
            )


def update_repository(database, repository, updated_sources, ssh_command):
    if ssh_command:
        os.environ["GIT_SSH_COMMAND"] = ssh_command
    image_repo = git.Repo(repository)
    if image_repo.is_dirty(untracked_files=True):
        print("\nChanges in local repository detected.\n")

        all_changes = []

        untracked_files = image_repo.untracked_files
        if untracked_files:
            print("New files:")
            for file in untracked_files:
                print(file)
                all_changes.append(file)

        changed_files = image_repo.git.diff(None, name_only=True)
        if changed_files:
            print("Updated files:")
            for file in changed_files.split("\n"):
                print(file)
                all_changes.append(file)

        releases_list = []

        for source in updated_sources:
            for release in updated_sources[source]["releases"]:
                release_data = db_get_release_versions(database, source, release, 1)
                releases_list.append(
                    release_data["name"] + " " + release_data["version"]
                )

        commit_message = "Added the following releases: " + ", ".join(releases_list)
        print("\n%s\n" % commit_message)

        image_repo.git.add(all_changes)
        image_repo.index.commit(commit_message)

    try:
        image_repo.remotes.origin.push()
    except Exception as error:
        print("ERROR: Push into upstream repository failed %s " % error)
